#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <dirent.h>

#include "utils.h"
#include "keys.h"
#include "polarssl/rsa.h"

#include "ctr.h"

// tw: bad code

void rehash_cro(char* fname, u8* hashtable_hash)
{
    u32 f_size, offset, len;
    u8 *cro;
    
    FILE* f = fopen(fname,"rb+");
    if(f == NULL)
    {
        printf("Failed to open %s!\n",fname);
        return;
    }
    fseek(f, 0, SEEK_END);
    f_size = ftell(f);
    cro = malloc(f_size);
    fseek(f, 0, SEEK_SET);
    fread(cro, 1, f_size, f);
    
    
    // Hash sections, write those to filebuf
    ctr_sha_256(cro + 0x80, 0x100, cro + 0x0);

    offset = getle32(cro + 0xB0);
    len = getle32(cro + 0xB4);
    ctr_sha_256(cro + offset, len, cro + 0x20);
    
    offset = getle32(cro + 0xC0);
    len = getle32(cro + 0xB8) - getle32(cro + 0xC0);
    ctr_sha_256(cro + offset, len, cro + 0x40);

    offset = getle32(cro + 0xB8);
    len = getle32(cro + 0xBC);
    ctr_sha_256(cro + offset, len, cro + 0x60);
    
    // get hashtable hash for return
    ctr_sha_256(cro, 0x80, hashtable_hash);
    
    fseek(f, 0, SEEK_SET);
    fwrite(cro, 1, f_size, f);
    printf("%s rehashed!\n", fname);
    
    fclose(f);
    free(cro);
}

u8 cro_hashtable_verify(u8* crr_hashtable, u32 num_cros, u8* cro_hash)
{
    u32 i;
    for(i = 0; i < num_cros; i++)
    {
        if(!memcmp(crr_hashtable + (i * 0x20), cro_hash, 0x20))
            return 1;
    }
    return 0;
}

void verify_cro(char* fname, u8* hashtable_hash)
{
    u32 ret, f_size, offset, len;
    u8 *cro;
    
    FILE* f = fopen(fname,"rb+");
    if(f == NULL)
    {
        printf("Failed to open %s!\n",fname);
        return;
    }
    fseek(f, 0, SEEK_END);
    f_size = ftell(f);
    cro = malloc(f_size);
    fseek(f, 0, SEEK_SET);
    fread(cro, 1, f_size, f);
    fclose(f);
    
    // Hash the hash table to return for verification
    ctr_sha_256(cro, 0x80, hashtable_hash);
    
    // Hash and verify the header
    ret = ctr_sha_256_verify(cro + 0x80, 0x100, cro);
    if(ret == 1)
        printf("%s header valid!\n",fname);
    else
        printf("%s header invalid!\n",fname);
    
    // Hash and verify sections
    offset = getle32(cro + 0xB0);
    len = getle32(cro + 0xB4);
    ret = ctr_sha_256_verify(cro + offset, len, cro + 0x20);
    if(ret == 1)
        printf("%s section 0 valid!\n",fname);
    else
        printf("%s section 0 invalid!\n",fname);
    
    offset = getle32(cro + 0xC0);
    len = getle32(cro + 0xB8) - getle32(cro + 0xC0);
    ret = ctr_sha_256_verify(cro + offset, len, cro + 0x40);
    if(ret == 1)
        printf("%s section 1 valid!\n",fname);
    else
        printf("%s section 1 invalid!\n",fname);

    offset = getle32(cro + 0xB8);
    len = getle32(cro + 0xBC);
    ret = ctr_sha_256_verify(cro + offset, len, cro + 0x60);
    if(ret == 1)
        printf("%s section 2 valid!\n",fname);
    else
        printf("%s section 2 invalid!\n",fname);
    
    free(cro);
}

int validate_crr(char* fname)
{
    
    rsakey2048 key;
    u32 ret, i, f_size, signed_area_size, num_cros;
    u8 *crr;
    u8 modulus[0x100];
    u8 sig[0x100];
    u8 hash[0x20];
    u8 cro_hashtable[0x80];
    
    FILE *f = fopen(fname,"rb+");
    if(f == NULL)
    {
        printf("Failed to open %s!\n",fname);
        return -1;
    }
    fseek(f, 0, SEEK_END);
    f_size = ftell(f);
    crr = malloc(f_size);
    fseek(f, 0, SEEK_SET);
    fread(crr, 1, f_size, f);
    
    // verify nintendo signature
    ctr_rsa_init_key_pubmodulus(&key, nin_rsa_mod);

    ctr_sha_256(crr + 0x20, 0x120, hash);
    
    ret = ctr_rsa_verify_hash(crr + 0x140, hash, &key);
    if(ret == 1)
        printf("Nintendo CRR RSA signature valid!\n\n");
    else
        printf("Nintendo CRR RSA signature invalid! (RSA patch needed)\n\n");
    
    // verify local signature
    ctr_rsa_init_key_pubmodulus(&key, crr + 0x40);

    signed_area_size = getle32(crr + 0x358) - 0x340;

    ctr_sha_256(crr + 0x340, signed_area_size, hash);
    
    ret = ctr_rsa_verify_hash(crr + 0x240, hash, &key);
    if(ret == 1)
        printf("Local CRR RSA signature valid!\n\n");
    else
        printf("Local CRR RSA signature invalid!\n\n");
    
    num_cros = getle32(crr + 0x354);

        DIR *dp;
    struct dirent *ep;
    char fnames[num_cros][0x100];
    i = -1;

    dp = opendir("./");
    if(dp != NULL)
    {
        while((ep = readdir(dp)))
        {
            if(strlen(ep->d_name) > 4 && !strcmp(ep->d_name + strlen(ep->d_name) - 4, ".cro"))
            {
                i++;
                if(i >= num_cros) {
                    printf("Invalid number of CROs! Expected %d, got %d+\n", num_cros, i+1);
                    return -1;
                }
                strncpy(fnames[i], ep->d_name, 0xFF);
            }
        }
        closedir(dp);
    }
    else {
        printf("Couldn't open the directory\n");
        return -1;
    }
        
    if(i+1 != num_cros) {
        printf("Invalid number of CROs! Expected %d, got %d\n", num_cros, i+1);
        return -1;
    }

    u8 hashtable_hash[0x20];
    
    for(i = 0; i < num_cros; i++) {
        verify_cro(fnames[i], hashtable_hash);
        ret = cro_hashtable_verify(crr + getle32(crr + 0x350), num_cros, hashtable_hash);
        if(ret) printf("%s hashtable valid!\n\n",fnames[i]);
        else printf("%s hashtable invalid!\n\n",fnames[i]);
    }

    free(crr);
    fclose(f);
    return 0;
}

void rsa_init_keys(rsa_context *ctx)
{
    mpi_read_binary(&ctx->N, rsa_mod, sizeof(rsa_mod));
    mpi_read_binary(&ctx->E, rsa_exponent, sizeof(rsa_exponent));
    //rsa_check_pubkey(&ctx);
    
    mpi_read_binary(&ctx->D, rsa_priv_exponent, sizeof(rsa_priv_exponent));
    mpi_read_binary(&ctx->P, rsa_priv_prime1, sizeof(rsa_priv_prime1));
    mpi_read_binary(&ctx->Q, rsa_priv_prime2, sizeof(rsa_priv_prime2));
    mpi_read_binary(&ctx->DP, rsa_priv_DP, sizeof(rsa_priv_DP));
    mpi_read_binary(&ctx->DQ, rsa_priv_DQ, sizeof(rsa_priv_DQ));
    mpi_read_binary(&ctx->QP, rsa_priv_QP, sizeof(rsa_priv_QP));
    //rsa_check_privkey(&ctx);
}

void rsa_sign_crr(u8* signature_out, u8* mod_out, u8* hash_data, u32 hash_len)
{
    rsa_context ctx;
    u8 hash[0x20];
    rsa_init(&ctx, RSA_PKCS_V15, 0);
    ctx.len = 0x100;
    
    ctr_sha_256(hash_data, hash_len, hash);
    
    rsa_init_keys(&ctx);
    
    rsa_pkcs1_sign(&ctx, RSA_PRIVATE, SIG_RSA_SHA256, 0x20, hash, signature_out);
    memcpy(mod_out, rsa_mod, sizeof(rsa_mod));
    rsa_free(&ctx);
}

int resign_crr(char* fname)
{
    u32 num_cros, f_size, i;
    u8* crr;
    
    FILE* f = fopen(fname,"rb+");
    if(f == NULL)
    {
        printf("Failed to open %s!\n",fname);
        return -1;
    }
    fseek(f, 0, SEEK_END);
    f_size = ftell(f);
    crr = malloc(f_size);
    fseek(f, 0, SEEK_SET);
    fread(crr, 1, f_size, f);
    
    num_cros = getle32(crr + 0x354);
    
    DIR *dp;
    struct dirent *ep;
    char fnames[num_cros][0x100];
    i = -1;

    dp = opendir("./");
    if(dp != NULL)
    {
        while((ep = readdir(dp)))
        {
            if(strlen(ep->d_name) > 4 && !strcmp(ep->d_name + strlen(ep->d_name) - 4, ".cro"))
            {
                i++;
                if(i >= num_cros) {
                    printf("Invalid number of CROs! Expected %d, got %d+\n", num_cros, i+1);
                    return -1;
                }
                strncpy(fnames[i], ep->d_name, 0xFF);
            }
        }
        closedir(dp);
    }
    else {
        printf("Couldn't open the directory\n");
        return -1;
    }
        
    if(i+1 != num_cros) {
        printf("Invalid number of CROs! Expected %d, got %d", num_cros, i+1);
        return -1;
    }
    
    u8* crr_hashtable = crr + getle32(crr + 0x350);
    
    for(i = 0; i < num_cros; i++)
        rehash_cro(fnames[i], crr_hashtable + (0x20 * i));
    printf("CRR rehashed!\n");
    
    rsa_sign_crr(crr + 0x240, crr + 0x40, crr + 0x340, getle32(crr + 0x358) - 0x340);
    printf("CRR resigned!\n");
    
    fseek(f, 0, SEEK_SET);
    fwrite(crr, 1, f_size, f);
    
    free(crr);
    fclose(f);
    return 0;
}

void usage()
{
    printf("cro_tool v1.2, by WulfyStylez\n");
    printf("Usage: cro_tool [-v] static.crr\n");
    printf("\t-v: verify CRR and CROs\n");
}

int main(int argc, char* argv[])
{
    if(argc < 2)
        usage();
    else if(!strcmp(argv[1],"-v"))
        validate_crr(argv[2]);
    else
        resign_crr(argv[1]);
    
	return 0;
}
